<!--
  ~ Copyright (c) Jupyter Development Team.
  ~ Distributed under the terms of the Modified BSD License.
-->

<!doctype html>
<html lang="en">
  <head>
    <title>Jupyter Services Demo</title>
    {% set page_config_full = {'baseUrl': base_url} %}

    <script id="jupyter-config-data" type="application/json">
      {{ page_config_full | tojson }}
    </script>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/require.js/2.2.0/require.js"></script>
    <style type="text/css">
    #cell {
      font-family: monospace;
    }
    </style>
  </head>
  <body>

    <h1>Run code!</h1>
    <p>
      Type code in the text area and click run to execute it.
      You will see a log of the messages produced by the kernel below.
    </p>
    <div>
      Kernel:
      <pre id="kernel-info">Starting…</pre>
    </div>
    <textarea id="cell" cols="100" rows="24"></textarea>
    <div>
      <button id="run">run</button>
    </div>
    <div id="output"></div>
    <script type="text/javascript">
    require.config({
        paths: {
          'jquery': 'https://code.jquery.com/jquery-2.2.4.min',
          '@jupyterlab/services': 'https://unpkg.com/@jupyterlab/services@*/dist/dist/index',
        },
        waitSeconds: 60
    });
    </script>
    <script src="{{page_config['fullStaticUrl'] | e}}/index.js" main="index"></script>
  </body>
</html>
