/*-----------------------------------------------------------------------------
| Copyright (c) Jupyter Development Team.
| Distributed under the terms of the Modified BSD License.
|----------------------------------------------------------------------------*/

/* Sibling imports */
@import './core.css';
@import './datagrid.css';
@import './dockpanel.css';
@import './menus.css';
@import './scrollbar.css';
@import './tabs.css';
@import './buttons.css';
@import './sidepanel.css';
@import './titlepanel.css';
@import './skiplink.css';
