/*-----------------------------------------------------------------------------
| Copyright (c) Jupyter Development Team.
| Distributed under the terms of the Modified BSD License.
|----------------------------------------------------------------------------*/

button.jp-Button.jp-mod-minimal.jp-TreeView.jp-TreeView,
button.jp-Button.jp-mod-minimal.jp-TreeView.jp-TableView {
  display: flex;
  align-items: center;
  -webkit-transition: 0.4s;
  transition: 0.4s;
  width: 35px;
}

.jp-ViewModeSelected .jp-Button,
.jp-PauseOnExceptions.lm-mod-toggled {
  background-color: var(--jp-inverse-layout-color3);
}

.jp-ViewModeSelected .jp-Button:hover,
.jp-PauseOnExceptions.lm-mod-toggled:hover {
  background-color: var(--jp-inverse-layout-color4);
}

.jp-ViewModeSelected .jp-Button .jp-icon3[fill],
.jp-PauseOnExceptions.lm-mod-toggled .jp-icon3[fill] {
  fill: var(--jp-layout-color1);
}
