/*-----------------------------------------------------------------------------
| Copyright (c) Jupyter Development Team.
| Distributed under the terms of the Modified BSD License.
|----------------------------------------------------------------------------*/

@import './variables.css';

/* Set the default typography for monospace elements */
tt,
code,
kbd,
samp,
pre {
  font-family: var(--jp-code-font-family);
  font-size: var(--jp-code-font-size);
  line-height: var(--jp-code-line-height);
}

/* fix illegible yellow backround in exception stacktrace */
:where(
    .jp-RenderedText[data-mime-type='application/vnd.jupyter.stderr']
      pre
      .ansi-yellow-bg
  ) {
  color: black;
}
